import java.util.*;
import java.io.*;

/**
 * A utility class that counts words in a text file
 * @author Bob Wilson (updated from Robert Cohen's pre-generic version)
 * 12/12/2009
 */
public class WordCount {

  /**
   * Delimiters separating words.
   */
  private static final String DELIMITERS = "[ \t\n\r\f,.()-?!;:&%@'\"]+";

  /**
   * Runs the application
   */
  public static void main(String[] args) throws IOException {
    if (args.length < 2) {
      printUsage();
    }
    String fileName = args[0];
    int count = 0;
    try {
      count = Integer.parseInt(args[1]);
    }
    catch (NumberFormatException e) {
      System.out.println(e.getMessage());
      printUsage();
    }
    Map<String,Counter> map = null;
    Scanner scan = null;
    
    // COMPLETE THIS LINE OF CODE THAT CREATES A File NAMED file TO READ
    // THE TEXTFILE WHOSE NAME IS FOUND IN THE VARIABLE String fileName
    

    // USE TRY-CATCH TO PROTECT THE FOLLOWING CODE AGAINST IOEXCEPTIONS
    scan = new Scanner(file);
    scan.useDelimiter(DELIMITERS);
    
    // NO NEED TO CHANGE ANYTHING AFTER HERE
    // COMPLEX JAVA GENERICS (WAIT UNTIL CS210)
                      
    List<Map.Entry<String,Counter>> result = filterByCount(orderByCount(countWords(scan)), count);
    if (result.size() == 0) {
      System.out.println("No words with frequency " + count + " or more.");
      return;
    }

    Iterator<Map.Entry<String,Counter>> entries = result.iterator();
    while (entries.hasNext()) {
      Map.Entry<String,Counter> entry = entries.next();
      String word = entry.getKey();
      Counter ctr = entry.getValue();
      System.out.println(word + '\t' + ctr);
    }
  }

  /**
   * Returns a map of the word frequencies.  The key of the returned map is the word
   * and the values are Counter objects of frequencies
   */
  private static Map<String,Counter> countWords(Scanner scan)
  {
    HashMap<String, Counter> map = new HashMap<String, Counter>();
    while(scan.hasNext())
    {
      String word = scan.next().toLowerCase();

      if (word.length() != 0) {
      Counter c = map.get(word);
      if (c == null)
        map.put(word, new Counter());
      else
        c.increment();
      }
    }
    return map;
  }

  /**
   * Takes a map with the key being the frequencies and counter being the
   * values, and returns a list of entries ordered by decreasing count.
   */
  private static List<Map.Entry<String,Counter>> orderByCount(Map<String,Counter> map) {
    ArrayList<Map.Entry<String,Counter>> result 
         =  new ArrayList<Map.Entry<String,Counter>>(map.entrySet());
    Collections.sort(result, new CounterValueComparator());
    return result;
  }

  /**
   * Takes a list ordered by decreasing count and returns a new list
   * of entries with count greater than or equal to a minimum count
   */
  private static List<Map.Entry<String,Counter>> filterByCount(List<Map.Entry<String,Counter>> l, int minCount) {
    ArrayList<Map.Entry<String,Counter>> ret = new ArrayList<Map.Entry<String,Counter>>(l);
    for (int i = ret.size() - 1; i >= 0; i--) {
      if ( ( (Counter) ret.get(i).getValue()).getValue() <
          minCount) {
        ret.remove(i);
      }
    }
    return ret;
  }

  /**
   * Prints a usage message
   */
  private static void printUsage() {
    System.out.println("Usage: java WordCount filename minCount");
    System.out.println("\tfilename is the name of the text file.");
    System.out
        .println("\tminCount is the minimum frequency of reported words.");
    System.exit(0);
  }

}
